router id 192.0.2.31;

log "/var/log/bird.log" all;
log syslog all;
debug protocols all;

protocol device { }

protocol bgp the_rs {
	local as 3;
	neighbor {{ data.rs_IPAddress }} as 999;
	next hop self;
	import all;
	export all;
	connect delay time 1;
	connect retry time 1;
}
