# built by ARouteServer
{{ "header"|include_local_file -}}

{% include "header.j2" %}


{% include "irrdb.j2" %}


{% include "clients.j2" %}


{% include "filters.j2" %}

{{ "footer"|include_local_file -}}
